import torch

from ..utils.decorators import ignore_numpy_warnings, default_internal
from ._shared_methods import select_target
from .psf_model_object import PSF_Model
from ..param import Param_Unlock, Param_SoftLimits

__all__ = ("Airy_PSF",)


class Airy_PSF(PSF_Model):
    """The Airy disk is an analytic description of the diffraction pattern
    for a circular aperture.

    The diffraction pattern is described exactly by the configuration
    of the lens system under the assumption that all elements are
    perfect. This expression goes as:

    .. math::

      I(\\theta) = I_0\\left[\\frac{2J_1(x)}{x}\\right]^2

      x = ka\\sin(\\theta) = \\frac{2\\pi a r}{\\lambda R}

    where :math:`I(\\theta)` is the intensity as a function of the
    angular position within the diffraction system along its main
    axis, :math:`I_0` is the central intensity of the airy disk,
    :math:`J_1` is the Bessel function of the first kind of order one,
    :math:`k = \\frac{2\\pi}{\\lambda}` is the wavenumber of the
    light, :math:`a` is the aperture radius, :math:`r` is the radial
    position from the center of the pattern, :math:`R` is the distance
    from the circular aperture to the observation plane.

    In the `Airy_PSF` class we combine the parameters
    :math:`a,R,\\lambda` into a single ratio to be optimized (or fixed
    by the optical configuration).

    """

    model_type = f"airy {PSF_Model.model_type}"
    parameter_specs = {
        "I0": {"units": "log10(flux/arcsec^2)", "value": 0.0, "locked": True},
        "aRL": {"units": "a/(R lambda)"},
    }
    _parameter_order = PSF_Model._parameter_order + ("I0", "aRL")
    usable = True
    model_integrated = False

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        if (parameters["I0"].value is not None) and (parameters["aRL"].value is not None):
            return
        target_area = target[self.window]
        icenter = target_area.plane_to_pixel(parameters["center"].value)

        if parameters["I0"].value is None:
            with Param_Unlock(parameters["I0"]), Param_SoftLimits(parameters["I0"]):
                parameters["I0"].value = torch.log10(
                    torch.mean(
                        target_area.data[
                            int(icenter[0]) - 2 : int(icenter[0]) + 2,
                            int(icenter[1]) - 2 : int(icenter[1]) + 2,
                        ]
                    )
                    / target.pixel_area.item()
                )
                parameters["I0"].uncertainty = torch.std(
                    target_area.data[
                        int(icenter[0]) - 2 : int(icenter[0]) + 2,
                        int(icenter[1]) - 2 : int(icenter[1]) + 2,
                    ]
                ) / (torch.abs(parameters["I0"].value) * target.pixel_area)
        if parameters["aRL"].value is None:
            with Param_Unlock(parameters["aRL"]), Param_SoftLimits(parameters["aRL"]):
                parameters["aRL"].value = (5.0 / 8.0) * 2 * target.pixel_length
                parameters["aRL"].uncertainty = parameters["aRL"].value * self.default_uncertainty

    @default_internal
    def radial_model(self, R, image=None, parameters=None):
        x = 2 * torch.pi * parameters["aRL"].value * R

        return (image.pixel_area * 10 ** parameters["I0"].value) * (
            2 * torch.special.bessel_j1(x) / x
        ) ** 2

    from ._shared_methods import radial_evaluate_model as evaluate_model



from typing import Optional, Union
import io
from copy import deepcopy

import numpy as np
import torch
from torch.autograd.functional import jacobian as torchjac

from ..param import Parameter_Node, Param_Mask
from ..utils.decorators import default_internal
from ..utils.interpolate import (
    _shift_Lanczos_kernel_torch,
    simpsons_kernel,
    curvature_kernel,
    interp2d,
)
from ..image import (
    Window,
    Jacobian_Image,
    Window_List,
    PSF_Image,
)
from ..utils.operations import (
    fft_convolve_torch,
    grid_integrate,
    single_quad_integrate,
)
from ..errors import SpecificationConflict
from .core_model import AstroPhot_Model
from .. import AP_config


@default_internal
def angular_metric(self, X, Y, image=None, parameters=None):
    return torch.atan2(Y, X)


@default_internal
def radius_metric(self, X, Y, image=None, parameters=None):
    return torch.sqrt(X**2 + Y**2 + self.softening**2)


@classmethod
def build_parameter_specs(cls, user_specs=None):
    parameter_specs = {}
    for base in cls.__bases__:
        try:
            parameter_specs.update(base.build_parameter_specs())
        except AttributeError:
            pass
    parameter_specs.update(cls.parameter_specs)
    parameter_specs = deepcopy(parameter_specs)
    if isinstance(user_specs, dict):
        for p in user_specs:
            # If the user supplied a parameter object subclass, simply use that as is
            if isinstance(user_specs[p], Parameter_Node):
                parameter_specs[p] = user_specs[p]
            elif isinstance(
                user_specs[p], dict
            ):  # if the user supplied parameter specifications, update the defaults
                parameter_specs[p].update(user_specs[p])
            else:
                parameter_specs[p]["value"] = user_specs[p]

    return parameter_specs


def build_parameters(self):
    for p in self.__class__._parameter_order:
        # skip if the parameter already exists
        if p in self.parameters:
            continue
        # If a parameter object is provided, simply use as-is
        if isinstance(self.parameter_specs[p], Parameter_Node):
            self.parameters.link(self.parameter_specs[p].to())
        elif isinstance(self.parameter_specs[p], dict):
            self.parameters.link(Parameter_Node(p, **self.parameter_specs[p]))
        else:
            raise ValueError(f"unrecognized parameter specification for {p}")


def _sample_init(self, image, parameters, center):
    if self.sampling_mode == "midpoint":
        Coords = image.get_coordinate_meshgrid()
        X, Y = Coords - center[..., None, None]
        mid = self.evaluate_model(X=X, Y=Y, image=image, parameters=parameters)
        kernel = curvature_kernel(AP_config.ap_dtype, AP_config.ap_device)
        # convolve curvature kernel to numericall compute second derivative
        curvature = torch.nn.functional.pad(
            torch.nn.functional.conv2d(
                mid.view(1, 1, *mid.shape),
                kernel.view(1, 1, *kernel.shape),
                padding="valid",
            ),
            (1, 1, 1, 1),
            mode="replicate",
        ).squeeze()
        return mid + curvature, mid
    elif self.sampling_mode == "simpsons":
        Coords = image.get_coordinate_simps_meshgrid()
        X, Y = Coords - center[..., None, None]
        dens = self.evaluate_model(X=X, Y=Y, image=image, parameters=parameters)
        kernel = simpsons_kernel(dtype=AP_config.ap_dtype, device=AP_config.ap_device)
        # midpoint is just every other sample in the simpsons grid
        mid = dens[1::2, 1::2]
        simps = torch.nn.functional.conv2d(
            dens.view(1, 1, *dens.shape), kernel, stride=2, padding="valid"
        )
        return mid.squeeze(), simps.squeeze()
    elif "quad" in self.sampling_mode:
        quad_level = int(self.sampling_mode[self.sampling_mode.find(":") + 1 :])
        Coords = image.get_coordinate_meshgrid()
        X, Y = Coords - center[..., None, None]
        res, ref = single_quad_integrate(
            X=X,
            Y=Y,
            image_header=image.header,
            eval_brightness=self.evaluate_model,
            eval_parameters=parameters,
            dtype=AP_config.ap_dtype,
            device=AP_config.ap_device,
            quad_level=quad_level,
        )
        return ref, res
    elif self.sampling_mode == "trapezoid":
        Coords = image.get_coordinate_corner_meshgrid()
        X, Y = Coords - center[..., None, None]
        dens = self.evaluate_model(X=X, Y=Y, image=image, parameters=parameters)
        kernel = (
            torch.ones((1, 1, 2, 2), dtype=AP_config.ap_dtype, device=AP_config.ap_device) / 4.0
        )
        trapz = torch.nn.functional.conv2d(dens.view(1, 1, *dens.shape), kernel, padding="valid")
        trapz = trapz.squeeze()
        kernel = curvature_kernel(AP_config.ap_dtype, AP_config.ap_device)
        curvature = torch.nn.functional.pad(
            torch.nn.functional.conv2d(
                trapz.view(1, 1, *trapz.shape),
                kernel.view(1, 1, *kernel.shape),
                padding="valid",
            ),
            (1, 1, 1, 1),
            mode="replicate",
        ).squeeze()
        return trapz + curvature, trapz

    raise SpecificationConflict(
        f"{self.name} has unknown sampling mode: {self.sampling_mode}. Should be one of: midpoint, simpsons, quad:level, trapezoid"
    )


def _integrate_reference(self, image_data, image_header, parameters):
    return torch.sum(image_data) / image_data.numel()


def _sample_integrate(self, deep, reference, image, parameters, center):
    if self.integrate_mode == "none":
        pass
    elif self.integrate_mode == "threshold":
        Coords = image.get_coordinate_meshgrid()
        X, Y = Coords - center[..., None, None]
        ref = self._integrate_reference(
            deep, image.header, parameters
        )  # fixme, error can be over 100% on initial sampling reference is invalid
        error = torch.abs((deep - reference))
        select = error > (self.sampling_tolerance * ref)
        intdeep = grid_integrate(
            X=X[select],
            Y=Y[select],
            image_header=image.header,
            eval_brightness=self.evaluate_model,
            eval_parameters=parameters,
            dtype=AP_config.ap_dtype,
            device=AP_config.ap_device,
            quad_level=self.integrate_quad_level,
            gridding=self.integrate_gridding,
            max_depth=self.integrate_max_depth,
            reference=self.sampling_tolerance * ref,
        )
        deep[select] = intdeep
    else:
        raise SpecificationConflict(
            f"{self.name} has unknown integration mode: {self.integrate_mode}. Should be one of: none, threshold"
        )
    return deep


def _shift_psf(self, psf, shift, shift_method="bilinear", keep_pad=True):
    if shift_method == "bilinear":
        psf_data = torch.nn.functional.pad(psf.data, (1, 1, 1, 1))
        X, Y = torch.meshgrid(
            torch.arange(
                psf_data.shape[1],
                dtype=AP_config.ap_dtype,
                device=AP_config.ap_device,
            )
            - shift[0],
            torch.arange(
                psf_data.shape[0],
                dtype=AP_config.ap_dtype,
                device=AP_config.ap_device,
            )
            - shift[1],
            indexing="xy",
        )
        shift_psf = interp2d(psf_data, X.clone(), Y.clone())
        if not keep_pad:
            shift_psf = shift_psf[1:-1, 1:-1]

    elif "lanczos" in shift_method:
        lanczos_order = int(shift_method[shift_method.find(":") + 1 :])
        psf_data = torch.nn.functional.pad(
            psf.data, (lanczos_order, lanczos_order, lanczos_order, lanczos_order)
        )
        LL = _shift_Lanczos_kernel_torch(
            -shift[0],
            -shift[1],
            lanczos_order,
            AP_config.ap_dtype,
            AP_config.ap_device,
        )
        shift_psf = torch.nn.functional.conv2d(
            psf_data.view(1, 1, *psf_data.shape),
            LL.view(1, 1, *LL.shape),
            padding="same",
        ).squeeze()
        if not keep_pad:
            shift_psf = shift_psf[lanczos_order:-lanczos_order, lanczos_order:-lanczos_order]
    else:
        raise SpecificationConflict(f"unrecognized subpixel shift method: {shift_method}")
    return shift_psf


def _sample_convolve(self, image, shift, psf, shift_method="bilinear"):
    """
    image: Image object with image.data pixel matrix
    shift: the amount of shifting to do in pixel units
    psf: a PSF_Image object
    """
    if shift is not None:
        shift_psf = self._shift_psf(psf, shift, shift_method)
    else:
        shift_psf = psf.data
    shift_psf = shift_psf / torch.sum(shift_psf)

    if self.psf_convolve_mode == "fft":
        image.data = fft_convolve_torch(image.data, shift_psf, img_prepadded=True)
    elif self.psf_convolve_mode == "direct":
        image.data = torch.nn.functional.conv2d(
            image.data.view(1, 1, *image.data.shape),
            torch.flip(
                shift_psf.view(1, 1, *shift_psf.shape),
                dims=(2, 3),
            ),
            padding="same",
        ).squeeze()
    else:
        raise ValueError(f"unrecognized psf_convolve_mode: {self.psf_convolve_mode}")


@torch.no_grad()
def jacobian(
    self,
    parameters: Optional[torch.Tensor] = None,
    as_representation: bool = False,
    window: Optional[Window] = None,
    pass_jacobian: Optional[Jacobian_Image] = None,
    **kwargs,
):
    """Compute the Jacobian matrix for this model.

    The Jacobian matrix represents the partial derivatives of the
    model's output with respect to its input parameters. It is useful
    in optimization and model fitting processes. This method
    simplifies the process of computing the Jacobian matrix for
    astronomical image models and is primarily used by the
    Levenberg-Marquardt algorithm for model fitting tasks.

    Args:
      parameters (Optional[torch.Tensor]): A 1D parameter tensor to override the
                                           current model's parameters.
      as_representation (bool): Indicates if the parameters argument is
                                provided as real values or representations
                                in the (-inf, inf) range. Default is False.
      parameters_identity (Optional[tuple]): Specifies which parameters are to be
                                             considered in the computation.
      window (Optional[Window]): A window object specifying the region of interest
                                 in the image.
      **kwargs: Additional keyword arguments.

    Returns:
      Jacobian_Image: A Jacobian_Image object containing the computed Jacobian matrix.

    """
    if window is None:
        window = self.window
    else:
        if isinstance(window, Window_List):
            window = window.window_list[pass_jacobian.index(self.target)]
        window = self.window & window

    # skip jacobian calculation if no parameters match criteria
    if torch.sum(self.parameters.vector_mask()) == 0 or window.overlap_frac(self.window) <= 0:
        return self.target[window].jacobian_image()

    # Set the parameters if provided and check the size of the parameter list
    if parameters is not None:
        if as_representation:
            self.parameters.vector_set_representation(parameters)
        else:
            self.parameters.vector_set_values(parameters)
    if torch.sum(self.parameters.vector_mask()) > self.jacobian_chunksize:
        return self._chunk_jacobian(
            as_representation=as_representation,
            window=window,
            **kwargs,
        )
    if torch.max(window.pixel_shape) > self.image_chunksize:
        return self._chunk_image_jacobian(
            as_representation=as_representation,
            window=window,
            **kwargs,
        )

    # Compute the jacobian
    full_jac = torchjac(
        lambda P: self(
            image=None,
            parameters=P,
            as_representation=as_representation,
            window=window,
        ).data,
        (
            self.parameters.vector_representation().detach()
            if as_representation
            else self.parameters.vector_values().detach()
        ),
        strategy="forward-mode",
        vectorize=True,
        create_graph=False,
    )

    # Store the jacobian as a Jacobian_Image object
    jac_img = self.target[window].jacobian_image(
        parameters=self.parameters.vector_identities(),
        data=full_jac,
    )
    return jac_img


@torch.no_grad()
def _chunk_image_jacobian(
    self,
    as_representation: bool = False,
    parameters_identity: Optional[tuple] = None,
    window: Optional[Window] = None,
    **kwargs,
):
    """Evaluates the Jacobian in smaller chunks to reduce memory usage.

    For models acting on large windows it can be prohibitive to build
    the full Jacobian in a single pass. Instead this function breaks
    the image into chunks as determined by `self.image_chunksize`
    evaluates the Jacobian only for the sub-images, it then builds up
    the full Jacobian as a separate tensor.

    This is for internal use and should be called by the
    `self.jacobian` function when appropriate.

    """

    pids = self.parameters.vector_identities()
    jac_img = self.target[window].jacobian_image(
        parameters=pids,
    )

    pixel_shape = window.pixel_shape.detach().cpu().numpy()
    Ncells = np.int64(np.round(np.ceil(pixel_shape / self.image_chunksize)))
    cellsize = np.int64(np.round(window.pixel_shape / Ncells))

    for nx in range(Ncells[0]):
        for ny in range(Ncells[1]):
            subwindow = window.copy()
            subwindow.crop_to_pixel(
                (
                    (cellsize[0] * nx, min(pixel_shape[0], cellsize[0] * (nx + 1))),
                    (cellsize[1] * ny, min(pixel_shape[1], cellsize[1] * (ny + 1))),
                )
            )
            jac_img += self.jacobian(
                parameters=None,
                as_representation=as_representation,
                window=subwindow,
                **kwargs,
            )

    return jac_img


@torch.no_grad()
def _chunk_jacobian(
    self,
    as_representation: bool = False,
    parameters_identity: Optional[tuple] = None,
    window: Optional[Window] = None,
    **kwargs,
):
    """Evaluates the Jacobian in small chunks to reduce memory usage.

    For models with many parameters it can be prohibitive to build the
    full Jacobian in a single pass. Instead this function breaks the
    list of parameters into chunks as determined by
    `self.jacobian_chunksize` evaluates the Jacobian only for those,
    it then builds up the full Jacobian as a separate tensor. This is
    for internal use and should be called by the `self.jacobian`
    function when appropriate.

    """
    pids = self.parameters.vector_identities()
    jac_img = self.target[window].jacobian_image(
        parameters=pids,
    )

    for ichunk in range(0, len(pids), self.jacobian_chunksize):
        mask = torch.zeros(len(pids), dtype=torch.bool, device=AP_config.ap_device)
        mask[ichunk : ichunk + self.jacobian_chunksize] = True
        with Param_Mask(self.parameters, mask):
            jac_img += self.jacobian(
                parameters=None,
                as_representation=as_representation,
                window=window,
                **kwargs,
            )

    return jac_img


def load(self, filename: Union[str, dict, io.TextIOBase] = "AstroPhot.yaml", new_name=None):
    """Used to load the model from a saved state.

    Sets the model window to the saved value and updates all
    parameters with the saved information. This overrides the
    current parameter settings.

    Args:
        filename: The source from which to load the model parameters. Can be a string (the name of the file on disc), a dictionary (formatted as if from self.get_state), or an io.TextIOBase (a file stream to load the file from).

    """
    state = AstroPhot_Model.load(filename)
    if new_name is None:
        new_name = state["name"]
    self.name = new_name
    # Use window saved state to initialize model window
    self.window = Window(**state["window"])
    # reassign target in case a target list was given
    self._target_identity = state["target_identity"]
    self.target = self.target
    # Set any attributes which were not default
    for key in self.track_attrs:
        if key in state:
            setattr(self, key, state[key])
    # Load the parameter group, this is handled by the parameter group object
    if isinstance(state["parameters"], Parameter_Node):
        self.parameters = state["parameters"]
    else:
        self.parameters = Parameter_Node(self.name, state=state["parameters"])
    # Move parameters to the appropriate device and dtype
    self.parameters.to(dtype=AP_config.ap_dtype, device=AP_config.ap_device)
    # Re-create the aux PSF model if there was one
    if "psf" in state:
        if state["psf"].get("type", "AstroPhot_Model") == "PSF_Image":
            self.psf = PSF_Image(state=state["psf"])
        else:
            print(state["psf"])
            state["psf"]["parameters"] = self.parameters[state["psf"]["name"]]
            self.set_aux_psf(
                AstroPhot_Model(
                    name=state["psf"]["name"],
                    filename=state["psf"],
                    target=self.target,
                )
            )
    return state



import numpy as np
import torch

from astropy.visualization import HistEqStretch, ImageNormalize
from matplotlib.patches import Polygon
import matplotlib
from scipy.stats import iqr

from ..models import Group_Model, PSF_Model
from ..image import Image_List, Window_List
from .. import AP_config
from ..utils.conversions.units import flux_to_sb
from .visuals import *


__all__ = ["target_image", "psf_image", "model_image", "residual_image", "model_window"]


def target_image(fig, ax, target, window=None, **kwargs):
    """
    This function is used to display a target image using the provided figure and axes.

    Args:
        fig (matplotlib.figure.Figure): The figure object in which the target image will be displayed.
        ax (matplotlib.axes.Axes): The axes object on which the target image will be plotted.
        target (Image or Image_List): The image or list of images to be displayed.
        window (Window, optional): The window through which the image is viewed. If `None`, the window of the
            provided `target` is used. Defaults to `None`.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        fig (matplotlib.figure.Figure): The figure object containing the displayed target image.
        ax (matplotlib.axes.Axes): The axes object containing the displayed target image.

    Note:
        If the `target` is an `Image_List`, this function will recursively call itself for each image in the list.
        The `window` parameter and `kwargs` are passed unchanged to each recursive call.
    """

    # recursive call for target image list
    if isinstance(target, Image_List):
        for i in range(len(target.image_list)):
            target_image(fig, ax[i], target.image_list[i], window=window, **kwargs)
        return fig, ax
    if window is None:
        window = target.window
    if kwargs.get("flipx", False):
        ax.invert_xaxis()
    target_area = target[window]
    dat = np.copy(target_area.data.detach().cpu().numpy())
    if target_area.has_mask:
        dat[target_area.mask.detach().cpu().numpy()] = np.nan
    X, Y = target_area.get_coordinate_corner_meshgrid()
    X = X.detach().cpu().numpy()
    Y = Y.detach().cpu().numpy()
    sky = np.nanmedian(dat)
    noise = iqr(dat[np.isfinite(dat)], rng=(16, 84)) / 2
    if noise == 0:
        noise = np.nanstd(dat)
    vmin = sky - 5 * noise
    vmax = sky + 5 * noise

    if kwargs.get("linear", False):
        im = ax.pcolormesh(
            X,
            Y,
            dat,
            cmap=cmap_grad,
        )
    else:
        im = ax.pcolormesh(
            X,
            Y,
            dat,
            cmap="Greys",
            norm=ImageNormalize(
                stretch=HistEqStretch(
                    dat[np.logical_and(dat <= (sky + 3 * noise), np.isfinite(dat))]
                ),
                clip=False,
                vmax=sky + 3 * noise,
                vmin=np.nanmin(dat),
            ),
        )

        im = ax.pcolormesh(
            X,
            Y,
            np.ma.masked_where(dat < (sky + 3 * noise), dat),
            cmap=cmap_grad,
            norm=matplotlib.colors.LogNorm(),
            clim=[sky + 3 * noise, None],
        )

    ax.axis("equal")
    ax.set_xlabel("Tangent Plane X [arcsec]")
    ax.set_ylabel("Tangent Plane Y [arcsec]")

    return fig, ax


@torch.no_grad()
def psf_image(
    fig,
    ax,
    psf,
    window=None,
    cmap_levels=None,
    flipx=False,
    **kwargs,
):
    if isinstance(psf, PSF_Model):
        psf = psf()
    # recursive call for target image list
    if isinstance(psf, Image_List):
        for i in range(len(psf.image_list)):
            psf_image(fig, ax[i], psf.image_list[i], window=window, **kwargs)
        return fig, ax

    if window is None:
        window = psf.window
    if flipx:
        ax.invert_xaxis()

    # cut out the requested window
    psf = psf[window]

    # Evaluate the model image
    X, Y = psf.get_coordinate_corner_meshgrid()
    X = X.detach().cpu().numpy()
    Y = Y.detach().cpu().numpy()
    psf = psf.data.detach().cpu().numpy()

    # Default kwargs for image
    imshow_kwargs = {
        "cmap": cmap_grad,
        "norm": matplotlib.colors.LogNorm(),  # "norm": ImageNormalize(stretch=LogStretch(), clip=False),
    }

    # Update with user provided kwargs
    imshow_kwargs.update(kwargs)

    # if requested, convert the continuous colourmap into discrete levels
    if cmap_levels is not None:
        imshow_kwargs["cmap"] = matplotlib.colors.ListedColormap(
            list(imshow_kwargs["cmap"](c) for c in np.linspace(0.0, 1.0, cmap_levels))
        )

    # Plot the image
    im = ax.pcolormesh(X, Y, psf, **imshow_kwargs)

    # Enforce equal spacing on x y
    ax.axis("equal")
    ax.set_xlabel("PSF X [arcsec]")
    ax.set_ylabel("PSF Y [arcsec]")

    return fig, ax


@torch.no_grad()
def model_image(
    fig,
    ax,
    model,
    sample_image=None,
    window=None,
    target=None,
    showcbar=True,
    target_mask=False,
    cmap_levels=None,
    flipx=False,
    magunits=True,
    sample_full_image=False,
    **kwargs,
):
    """
    This function is used to generate a model image and display it using the provided figure and axes.

    Args:
        fig (matplotlib.figure.Figure): The figure object in which the image will be displayed.
        ax (matplotlib.axes.Axes): The axes object on which the image will be plotted.
        model (Model): The model object used to generate a model image if `sample_image` is not provided.
        sample_image (Image or Image_List, optional): The image or list of images to be displayed.
            If `None`, a model image is generated using the provided `model`. Defaults to `None`.
        window (Window, optional): The window through which the image is viewed. If `None`, the window of the
            provided `model` is used. Defaults to `None`.
        target (Target, optional): The target or list of targets for the image or image list.
            If `None`, the target of the `model` is used. Defaults to `None`.
        showcbar (bool, optional): Whether to show the color bar. Defaults to `True`.
        target_mask (bool, optional): Whether to apply the mask of the target. If `True` and if the target has a mask,
            the mask is applied to the image. Defaults to `False`.
        cmap_levels (int, optional): The number of discrete levels to convert the continuous color map to.
            If not `None`, the color map is converted to a ListedColormap with the specified number of levels.
            Defaults to `None`.
        sample_full_image: If True, every model will be sampled on the full image window. If False (default) each model will only be sampled in its fitting window.
        **kwargs: Arbitrary keyword arguments. These are used to override the default imshow_kwargs.

    Returns:
        fig (matplotlib.figure.Figure): The figure object containing the displayed image.
        ax (matplotlib.axes.Axes): The axes object containing the displayed image.

    Note:
        If the `sample_image` is an `Image_List`, this function will recursively call itself for each image in the list,
        with the corresponding target and window. The `showcbar` parameter and `kwargs` are passed unchanged to each recursive call.
    """

    if sample_image is None:
        if sample_full_image:
            sample_image = model.make_model_image()
            sample_image = model(sample_image)
        else:
            sample_image = model()

    # Use model target if not given
    if target is None:
        target = model.target

    # Use model window if not given
    if window is None:
        window = model.window

    # Handle image lists
    if isinstance(sample_image, Image_List):
        for i, images in enumerate(zip(sample_image, target, window)):
            model_image(
                fig,
                ax[i],
                model,
                sample_image=images[0],
                window=images[2],
                target=images[1],
                showcbar=showcbar,
                target_mask=target_mask,
                cmap_levels=cmap_levels,
                flipx=flipx,
                magunits=magunits,
                **kwargs,
            )
        return fig, ax

    if flipx:
        ax.invert_xaxis()

    # cut out the requested window
    sample_image = sample_image[window]

    # Evaluate the model image
    X, Y = sample_image.get_coordinate_corner_meshgrid()
    X = X.detach().cpu().numpy()
    Y = Y.detach().cpu().numpy()
    sample_image = sample_image.data.detach().cpu().numpy()

    # Default kwargs for image
    imshow_kwargs = {
        "cmap": cmap_grad,
        "norm": matplotlib.colors.LogNorm(),  # "norm": ImageNormalize(stretch=LogStretch(), clip=False),
    }

    # Update with user provided kwargs
    imshow_kwargs.update(kwargs)

    # if requested, convert the continuous colourmap into discrete levels
    if cmap_levels is not None:
        imshow_kwargs["cmap"] = matplotlib.colors.ListedColormap(
            list(imshow_kwargs["cmap"](c) for c in np.linspace(0.0, 1.0, cmap_levels))
        )

    # If zeropoint is available, convert to surface brightness units
    if target.zeropoint is not None and magunits:
        sample_image = flux_to_sb(sample_image, target.pixel_area.item(), target.zeropoint.item())
        del imshow_kwargs["norm"]
        imshow_kwargs["cmap"] = imshow_kwargs["cmap"].reversed()

    # Apply the mask if available
    if target_mask and target.has_mask:
        sample_image[target.mask.detach().cpu().numpy()] = np.nan

    # Plot the image
    im = ax.pcolormesh(X, Y, sample_image, **imshow_kwargs)

    # Enforce equal spacing on x y
    ax.axis("equal")
    ax.set_xlabel("Tangent Plane X [arcsec]")
    ax.set_ylabel("Tangent Plane Y [arcsec]")

    # Add a colourbar
    if showcbar:
        if target.zeropoint is not None and magunits:
            clb = fig.colorbar(im, ax=ax, label="Surface Brightness [mag/arcsec$^2$]")
            clb.ax.invert_yaxis()
        else:
            clb = fig.colorbar(im, ax=ax, label="log$_{10}$(flux)")

    return fig, ax


@torch.no_grad()
def residual_image(
    fig,
    ax,
    model,
    target=None,
    sample_image=None,
    showcbar=True,
    window=None,
    center_residuals=False,
    clb_label=None,
    normalize_residuals=False,
    flipx=False,
    sample_full_image=False,
    **kwargs,
):
    """
    This function is used to calculate and display the residuals of a model image with respect to a target image.
    The residuals are calculated as the difference between the target image and the sample image.

    Args:
        fig (matplotlib.figure.Figure): The figure object in which the residuals will be displayed.
        ax (matplotlib.axes.Axes): The axes object on which the residuals will be plotted.
        model (Model): The model object used to generate a model image if `sample_image` is not provided.
        target (Target or Image_List, optional): The target or list of targets for the image or image list.
            If `None`, the target of the `model` is used. Defaults to `None`.
        sample_image (Image or Image_List, optional): The image or list of images from which residuals will be calculated.
            If `None`, a model image is generated using the provided `model`. Defaults to `None`.
        showcbar (bool, optional): Whether to show the color bar. Defaults to `True`.
        window (Window or Window_List, optional): The window through which the image is viewed. If `None`, the window of the
            provided `model` is used. Defaults to `None`.
        center_residuals (bool, optional): Whether to subtract the median of the residuals. If `True`, the median is subtracted
            from the residuals. Defaults to `False`.
        clb_label (str, optional): The label for the colorbar. If `None`, a default label is used based on the normalization of the
            residuals. Defaults to `None`.
        normalize_residuals (bool, optional): Whether to normalize the residuals. If `True`, residuals are divided by the square root
            of the variance of the target. Defaults to `False`.
        sample_full_image: If True, every model will be sampled on the full image window. If False (default) each model will only be sampled in its fitting window.
        **kwargs: Arbitrary keyword arguments. These are used to override the default imshow_kwargs.

    Returns:
        fig (matplotlib.figure.Figure): The figure object containing the displayed residuals.
        ax (matplotlib.axes.Axes): The axes object containing the displayed residuals.

    Note:
        If the `window`, `target`, or `sample_image` are lists, this function will recursively call itself for each element in the list,
        with the corresponding window, target, and sample image. The `showcbar`, `center_residuals`, and `kwargs` are passed unchanged to
        each recursive call.
    """

    if window is None:
        window = model.window
    if target is None:
        target = model.target
    if sample_image is None:
        if sample_full_image:
            sample_image = model.make_model_image()
            sample_image = model(sample_image)
        else:
            sample_image = model()
    if isinstance(window, Window_List) or isinstance(target, Image_List):
        for i_ax, win, tar, sam in zip(ax, window, target, sample_image):
            residual_image(
                fig,
                i_ax,
                model,
                target=tar,
                sample_image=sam,
                window=win,
                showcbar=showcbar,
                center_residuals=center_residuals,
                clb_label=clb_label,
                normalize_residuals=normalize_residuals,
                flipx=flipx,
                **kwargs,
            )
        return fig, ax

    if flipx:
        ax.invert_xaxis()
    X, Y = sample_image[window].get_coordinate_corner_meshgrid()
    X = X.detach().cpu().numpy()
    Y = Y.detach().cpu().numpy()
    residuals = (target[window] - sample_image[window]).data
    if isinstance(normalize_residuals, bool) and normalize_residuals:
        residuals = residuals / torch.sqrt(target[window].variance)
    elif isinstance(normalize_residuals, torch.Tensor):
        residuals = residuals / torch.sqrt(normalize_residuals)
        normalize_residuals = True
    residuals = residuals.detach().cpu().numpy()

    if target.has_mask:
        residuals[target[window].mask.detach().cpu().numpy()] = np.nan
    if center_residuals:
        residuals -= np.nanmedian(residuals)
    residuals = np.arctan(residuals / (iqr(residuals[np.isfinite(residuals)], rng=[10, 90]) * 2))
    extreme = np.max(np.abs(residuals[np.isfinite(residuals)]))
    imshow_kwargs = {
        "cmap": cmap_div,
        "vmin": -extreme,
        "vmax": extreme,
    }
    imshow_kwargs.update(kwargs)
    im = ax.pcolormesh(X, Y, residuals, **imshow_kwargs)
    ax.axis("equal")
    ax.set_xlabel("Tangent Plane X [arcsec]")
    ax.set_ylabel("Tangent Plane Y [arcsec]")

    if showcbar:
        if normalize_residuals:
            default_label = f"tan$^{{-1}}$((Target - {model.name}) / $\\sigma$)"
        else:
            default_label = f"tan$^{{-1}}$(Target - {model.name})"
        clb = fig.colorbar(im, ax=ax, label=default_label if clb_label is None else clb_label)
        clb.ax.set_yticks([])
        clb.ax.set_yticklabels([])
    return fig, ax


def model_window(fig, ax, model, target=None, rectangle_linewidth=2, **kwargs):
    if isinstance(ax, np.ndarray):
        for i, axitem in enumerate(ax):
            model_window(fig, axitem, model, target=model.target.image_list[i], **kwargs)
        return fig, ax

    if isinstance(model, Group_Model):
        for m in model.models.values():
            if isinstance(m.window, Window_List):
                use_window = m.window.window_list[m.target.index(target)]
            else:
                use_window = m.window

            lowright = use_window.pixel_shape.clone().to(dtype=AP_config.ap_dtype)
            lowright[1] = 0.0
            lowright = use_window.origin + use_window.pixel_to_plane_delta(lowright)
            lowright = lowright.detach().cpu().numpy()
            upleft = use_window.pixel_shape.clone().to(dtype=AP_config.ap_dtype)
            upleft[0] = 0.0
            upleft = use_window.origin + use_window.pixel_to_plane_delta(upleft)
            upleft = upleft.detach().cpu().numpy()
            end = use_window.origin + use_window.end
            end = end.detach().cpu().numpy()
            x = [
                use_window.origin[0].detach().cpu().numpy(),
                lowright[0],
                end[0],
                upleft[0],
            ]
            y = [
                use_window.origin[1].detach().cpu().numpy(),
                lowright[1],
                end[1],
                upleft[1],
            ]
            ax.add_patch(
                Polygon(
                    xy=list(zip(x, y)),
                    fill=False,
                    linewidth=rectangle_linewidth,
                    edgecolor=main_pallet["secondary1"],
                )
            )
    else:
        if isinstance(model.window, Window_List):
            use_window = model.window.window_list[model.target.index(target)]
        else:
            use_window = model.window
        lowright = use_window.pixel_shape.clone().to(dtype=AP_config.ap_dtype)
        lowright[1] = 0.0
        lowright = use_window.origin + use_window.pixel_to_plane_delta(lowright)
        lowright = lowright.detach().cpu().numpy()
        upleft = use_window.pixel_shape.clone().to(dtype=AP_config.ap_dtype)
        upleft[0] = 0.0
        upleft = use_window.origin + use_window.pixel_to_plane_delta(upleft)
        upleft = upleft.detach().cpu().numpy()
        end = use_window.origin + use_window.end
        end = end.detach().cpu().numpy()
        x = [
            use_window.origin[0].detach().cpu().numpy(),
            lowright[0],
            end[0],
            upleft[0],
        ]
        y = [
            use_window.origin[1].detach().cpu().numpy(),
            lowright[1],
            end[1],
            upleft[1],
        ]
        ax.add_patch(
            Polygon(
                xy=list(zip(x, y)),
                fill=False,
                linewidth=rectangle_linewidth,
                edgecolor=main_pallet["secondary1"],
            )
        )

    return fig, ax



import numpy as np


def Rscale_Fmodes(theta, modes, Am, Phim):
    """Factor to scale radius values given a set of fourier mode
    amplitudes.

    """
    return np.exp(sum(Am[m] * np.cos(modes[m] * (theta + Phim[m])) for m in range(len(modes))))


def parametric_Fmodes(theta, modes, Am, Phim):
    """determines a number of scaled radius samples with fourier mode
    perturbations for a unit circle.

    """
    x = np.cos(theta)
    y = np.sin(theta)
    Rscale = Rscale_Fmodes(theta, modes, Am, Phim)
    return x * Rscale, y * Rscale


def Rscale_SuperEllipse(theta, ellip, C=2):
    """Scale factor for radius values given a super ellipse coefficient."""
    res = (1 - ellip) / (
        np.abs((1 - ellip) * np.cos(theta)) ** (C) + np.abs(np.sin(theta)) ** (C)
    ) ** (1.0 / C)
    return res


def parametric_SuperEllipse(theta, ellip, C=2):
    """determines a number of scaled radius samples with super ellipse
    perturbations for a unit circle.

    """
    rs = Rscale_SuperEllipse(theta, ellip, C)
    return rs * np.cos(theta), rs * np.sin(theta)



import torch
from astropy.wcs import WCS as AstropyWCS

from .. import AP_config
from .wcs import WCS
from ..errors import ConflicingWCS, SpecificationConflict

__all__ = ["Window", "Window_List"]


class Window(WCS):
    """class to define a window on the sky in coordinate space. These
    windows can undergo arithmetic and preserve logical behavior. Image
    objects can also be indexed using windows and will return an
    appropriate subsection of their data.

    There are several ways to tell a Window object where to
    place itself. The simplest method is to pass an
    Astropy WCS object such as::

      H = ap.image.Window(wcs = wcs)

    this will automatically place your image at the correct RA, DEC
    and assign the correct pixel scale, etc. WARNING, it will default to
    setting the reference RA DEC at the reference RA DEC of the wcs
    object; if you have multiple images you should force them all to
    have the same reference world coordinate by passing
    ``reference_radec = (ra, dec)``. See the :doc:`coordinates`
    documentation for more details. There are several other ways to
    initialize a window. If you provide ``origin_radec`` then
    it will place the image origin at the requested RA DEC
    coordinates. If you provide ``center_radec`` then it will place
    the image center at the requested RA DEC coordinates. Note that in
    these cases the fixed point between the pixel grid and image plane
    is different (pixel origin and center respectively); so if you
    have rotated pixels in your pixel scale matrix then everything
    will be rotated about different points (pixel origin and center
    respectively). If you provide ``origin`` or ``center`` then those
    are coordinates in the tangent plane (arcsec) and they will
    correspondingly become fixed points. For arbitrary control over
    the pixel positioning, use ``reference_imageij`` and
    ``reference_imagexy`` to fix the pixel and tangent plane
    coordinates respectively to each other, any rotation or shear will
    happen about that fixed point.

    Args:
      origin : Sequence or None, optional
          The origin of the image in the tangent plane coordinate system
          (arcsec), as a 1D array of length 2. Default is None.
      origin_radec : Sequence or None, optional
          The origin of the image in the world coordinate system (RA,
          DEC in degrees), as a 1D array of length 2. Default is None.
      center : Sequence or None, optional
          The center of the image in the tangent plane coordinate system
          (arcsec), as a 1D array of length 2. Default is None.
      center_radec : Sequence or None, optional
          The center of the image in the world coordinate system (RA,
          DEC in degrees), as a 1D array of length 2. Default is None.
      wcs: An astropy.wcs.WCS object which gives information about the
          origin and orientation of the window.
      reference_radec: world coordinates on the celestial sphere (RA,
          DEC in degrees) where the tangent plane makes contact. This should
          be the same for every image in multi-image analysis.
      reference_planexy: tangent plane coordinates (arcsec) where it
          makes contact with the celesial sphere. This should typically be
          (0,0) though that is not stricktly enforced (it is assumed if not
          given). This reference coordinate should be the same for all
          images in multi-image analysis.
      reference_imageij: pixel coordinates about which the image is
          defined. For example in an Astropy WCS object the wcs.wcs.crpix
          array gives the pixel coordinate reference point for which the
          world coordinate mapping (wcs.wcs.crval) is defined. One may think
          of the referenced pixel location as being "pinned" to the tangent
          plane. This may be different for each image in multi-image
          analysis..
      reference_imagexy: tangent plane coordinates (arcsec) about
          which the image is defined. This is the pivot point about which the
          pixelscale matrix operates, therefore if the pixelscale matrix
          defines a rotation then this is the coordinate about which the
          rotation will be performed. This may be different for each image in
          multi-image analysis.

    """

    def __init__(
        self,
        *,
        pixel_shape=None,
        origin=None,
        origin_radec=None,
        center=None,
        center_radec=None,
        state=None,
        fits_state=None,
        wcs=None,
        **kwargs,
    ):
        # If loading from a previous state, simply update values and end init
        if state is not None:
            self.set_state(state)
            return
        if fits_state is not None:
            self.set_fits_state(fits_state)
            return

        # Collect the shape of the window
        if pixel_shape is not None:
            self.pixel_shape = pixel_shape
        else:
            self.pixel_shape = wcs.pixel_shape

        # Determine relative positioning of tangent plane and pixel grid. Also world coordinates and tangent plane
        if not sum(C is not None for C in [wcs, origin_radec, center_radec, origin, center]) <= 1:
            raise SpecificationConflict(
                "Please provide only one reference position for the window, otherwise the placement is ambiguous"
            )

        # Image coordinates provided by WCS
        if wcs is not None:
            super().__init__(wcs=wcs, **kwargs)
        # Image reference position from RA and DEC of image origin
        elif origin_radec is not None:
            # Origin given, it is reference point
            origin_radec = torch.as_tensor(
                origin_radec, dtype=AP_config.ap_dtype, device=AP_config.ap_device
            )
            kwargs["reference_radec"] = kwargs.get("reference_radec", origin_radec)
            super().__init__(**kwargs)
            self.reference_imageij = (-0.5, -0.5)
            self.reference_imagexy = self.world_to_plane(origin_radec)
        # Image reference position from RA and DEC of image center
        elif center_radec is not None:
            pix_center = self.pixel_shape.to(dtype=AP_config.ap_dtype) / 2 - 0.5
            center_radec = torch.as_tensor(
                center_radec, dtype=AP_config.ap_dtype, device=AP_config.ap_device
            )
            kwargs["reference_radec"] = kwargs.get("reference_radec", center_radec)
            super().__init__(**kwargs)
            center = self.world_to_plane(center_radec)
            self.reference_imageij = pix_center
            self.reference_imagexy = center
        # Image reference position from tangent plane position of image origin
        elif origin is not None:
            kwargs.update(
                {
                    "reference_imageij": (-0.5, -0.5),
                    "reference_imagexy": origin,
                }
            )
            super().__init__(**kwargs)
        # Image reference position from tangent plane position of image center
        elif center is not None:
            pix_center = self.pixel_shape.to(dtype=AP_config.ap_dtype) / 2 - 0.5
            kwargs.update(
                {
                    "reference_imageij": pix_center,
                    "reference_imagexy": center,
                }
            )
            super().__init__(**kwargs)
        # Image origin assumed to be at tangent plane origin
        else:
            super().__init__(**kwargs)

    @property
    def shape(self):
        dtype, device = self.pixelscale.dtype, self.pixelscale.device
        S1 = self.pixel_shape.to(dtype=dtype, device=device)
        S1[1] = 0.0
        S2 = self.pixel_shape.to(dtype=dtype, device=device)
        S2[0] = 0.0
        return torch.stack(
            (
                torch.linalg.norm(self.pixelscale @ S1),
                torch.linalg.norm(self.pixelscale @ S2),
            )
        )

    @shape.setter
    def shape(self, shape):
        if shape is None:
            self._pixel_shape = None
            return
        shape = torch.as_tensor(shape, dtype=self.pixelscale.dtype, device=self.pixelscale.device)
        self.pixel_shape = shape / torch.sqrt(torch.sum(self.pixelscale**2, dim=0))

    @property
    def pixel_shape(self):
        return self._pixel_shape

    @pixel_shape.setter
    def pixel_shape(self, shape):
        if shape is None:
            self._pixel_shape = None
            return
        self._pixel_shape = torch.as_tensor(shape, device=AP_config.ap_device)
        self._pixel_shape = torch.round(self.pixel_shape).to(
            dtype=torch.int32, device=AP_config.ap_device
        )

    @property
    def size(self):
        """The number of pixels in the window"""
        return torch.prod(self.pixel_shape)

    @property
    def end(self):
        return self.pixel_to_plane_delta(
            self.pixel_shape.to(dtype=self.pixelscale.dtype, device=self.pixelscale.device)
        )

    @property
    def origin(self):
        return self.pixel_to_plane(-0.5 * torch.ones_like(self.reference_imageij))

    @property
    def center(self):
        return self.origin + self.end / 2

    def copy(self, **kwargs):
        copy_kwargs = {"pixel_shape": torch.clone(self.pixel_shape)}
        copy_kwargs.update(kwargs)
        return super().copy(**copy_kwargs)

    def to(self, dtype=None, device=None):
        if dtype is None:
            dtype = AP_config.ap_dtype
        if device is None:
            device = AP_config.ap_device
        super().to(dtype=dtype, device=device)
        self.pixel_shape = self.pixel_shape.to(dtype=dtype, device=device)

    def rescale_pixel(self, scale, **kwargs):
        return self.copy(
            pixelscale=self.pixelscale * scale,
            pixel_shape=self.pixel_shape // scale,
            reference_imageij=(self.reference_imageij + 0.5) / scale - 0.5,
            **kwargs,
        )

    @staticmethod
    @torch.no_grad()
    def _get_indices(ref_window, obj_window):
        other_origin_pix = torch.round(ref_window.plane_to_pixel(obj_window.origin) + 0.5).int()
        new_origin_pix = torch.maximum(torch.zeros_like(other_origin_pix), other_origin_pix)

        other_pixel_end = torch.round(
            ref_window.plane_to_pixel(obj_window.origin + obj_window.end) + 0.5
        ).int()
        new_pixel_end = torch.minimum(ref_window.pixel_shape, other_pixel_end)
        return slice(new_origin_pix[1], new_pixel_end[1]), slice(
            new_origin_pix[0], new_pixel_end[0]
        )

    def get_self_indices(self, obj):
        """
        Return an index slicing tuple for obj corresponding to this window
        """
        if isinstance(obj, Window):
            return self._get_indices(self, obj)
        return self._get_indices(self, obj.window)

    def get_other_indices(self, obj):
        """
        Return an index slicing tuple for obj corresponding to this window
        """
        if isinstance(obj, Window):
            return self._get_indices(obj, self)
        return self._get_indices(obj.window, self)

    def overlap_frac(self, other):
        overlap = self & other
        overlap_area = torch.prod(overlap.shape)
        full_area = torch.prod(self.shape) + torch.prod(other.shape) - overlap_area
        return overlap_area / full_area

    def shift(self, shift):
        """
        Shift the location of the window by a specified amount in tangent plane coordinates
        """
        self.reference_imagexy = self.reference_imagexy + shift
        return self

    def pixel_shift(self, shift):
        """
        Shift the location of the window by a specified amount in pixel grid coordinates
        """

        self.reference_imageij = self.reference_imageij - shift
        return self

    def get_astropywcs(self, **kwargs):
        wargs = {
            "NAXIS": 2,
            "NAXIS1": self.pixel_shape[0].item(),
            "NAXIS2": self.pixel_shape[1].item(),
            "CTYPE1": "RA---TAN",
            "CTYPE2": "DEC--TAN",
            "CRVAL1": self.pixel_to_world(self.reference_imageij)[0].item(),
            "CRVAL2": self.pixel_to_world(self.reference_imageij)[1].item(),
            "CRPIX1": self.reference_imageij[0].item(),
            "CRPIX2": self.reference_imageij[1].item(),
            "CD1_1": self.pixelscale[0][0].item(),
            "CD1_2": self.pixelscale[0][1].item(),
            "CD2_1": self.pixelscale[1][0].item(),
            "CD2_2": self.pixelscale[1][1].item(),
        }
        wargs.update(kwargs)
        return AstropyWCS(wargs)

    def get_state(self):
        state = super().get_state()
        state["pixel_shape"] = self.pixel_shape.detach().cpu().tolist()
        return state

    def set_state(self, state):
        super().set_state(state)
        self.pixel_shape = torch.tensor(
            state["pixel_shape"], dtype=AP_config.ap_dtype, device=AP_config.ap_device
        )

    def get_fits_state(self):
        state = super().get_fits_state()
        state["PXL_SHPE"] = str(self.pixel_shape.detach().cpu().tolist())
        return state

    def set_fits_state(self, state):
        super().set_fits_state(state)
        self.pixel_shape = torch.tensor(
            eval(state["PXL_SHPE"]),
            dtype=AP_config.ap_dtype,
            device=AP_config.ap_device,
        )

    def crop_pixel(self, pixels):
        """
        [crop all sides] or
        [crop x, crop y] or
        [crop x low, crop y low, crop x high, crop y high]
        """
        if len(pixels) == 1:
            self.pixel_shape = self.pixel_shape - 2 * pixels[0]
            self.reference_imageij = self.reference_imageij - pixels[0]
        elif len(pixels) == 2:
            pix_shift = torch.as_tensor(
                pixels, dtype=AP_config.ap_dtype, device=AP_config.ap_device
            )
            self.pixel_shape = self.pixel_shape - 2 * pix_shift
            self.reference_imageij = self.reference_imageij - pix_shift
        elif len(pixels) == 4:  # different crop on all sides
            pixels = torch.as_tensor(pixels, dtype=AP_config.ap_dtype, device=AP_config.ap_device)
            self.pixel_shape = self.pixel_shape - pixels[:2] - pixels[2:]
            self.reference_imageij = self.reference_imageij - pixels[:2]
        else:
            raise ValueError(f"Unrecognized pixel crop format: {pixels}")
        return self

    def crop_to_pixel(self, pixels):
        """
        format: [[xmin, xmax],[ymin,ymax]]
        """
        pixels = torch.tensor(pixels, dtype=AP_config.ap_dtype, device=AP_config.ap_device)
        self.reference_imageij = self.reference_imageij - pixels[:, 0]
        self.pixel_shape = pixels[:, 1] - pixels[:, 0]
        return self

    def pad_pixel(self, pixels):
        """
        [pad all sides] or
        [pad x, pad y] or
        [pad x low, pad y low, pad x high, pad y high]
        """
        if len(pixels) == 1:
            self.pixel_shape = self.pixel_shape + 2 * pixels[0]
            self.reference_imageij = self.reference_imageij + pixels[0]
        elif len(pixels) == 2:
            pix_shift = torch.as_tensor(
                pixels, dtype=AP_config.ap_dtype, device=AP_config.ap_device
            )
            self.pixel_shape = self.pixel_shape + 2 * pix_shift
            self.reference_imageij = self.reference_imageij + pix_shift
        elif len(pixels) == 4:  # different crop on all sides
            pixels = torch.as_tensor(pixels, dtype=AP_config.ap_dtype, device=AP_config.ap_device)
            self.pixel_shape = self.pixel_shape + pixels[:2] + pixels[2:]
            self.reference_imageij = self.reference_imageij + pixels[:2]
        else:
            raise ValueError(f"Unrecognized pixel crop format: {pixels}")
        return self

    @torch.no_grad()
    def get_coordinate_meshgrid(self):
        """Returns a meshgrid with tangent plane coordinates for the center
        of every pixel.

        """
        pix = self.pixel_shape.to(dtype=AP_config.ap_dtype)
        xsteps = torch.arange(pix[0], dtype=AP_config.ap_dtype, device=AP_config.ap_device)
        ysteps = torch.arange(pix[1], dtype=AP_config.ap_dtype, device=AP_config.ap_device)
        meshx, meshy = torch.meshgrid(
            xsteps,
            ysteps,
            indexing="xy",
        )
        Coords = self.pixel_to_plane(meshx, meshy)
        return torch.stack(Coords)

    @torch.no_grad()
    def get_coordinate_corner_meshgrid(self):
        """Returns a meshgrid with tangent plane coordinates for the corners
        of every pixel.

        """
        pix = self.pixel_shape.to(dtype=AP_config.ap_dtype)
        xsteps = (
            torch.arange(pix[0] + 1, dtype=AP_config.ap_dtype, device=AP_config.ap_device) - 0.5
        )
        ysteps = (
            torch.arange(pix[1] + 1, dtype=AP_config.ap_dtype, device=AP_config.ap_device) - 0.5
        )
        meshx, meshy = torch.meshgrid(
            xsteps,
            ysteps,
            indexing="xy",
        )
        Coords = self.pixel_to_plane(meshx, meshy)
        return torch.stack(Coords)

    @torch.no_grad()
    def get_coordinate_simps_meshgrid(self):
        """Returns a meshgrid with tangent plane coordinates for performing
        simpsons method pixel integration (all corners, centers, and
        middle of each edge). This is approximately 4 times more
        points than the standard :meth:`get_coordinate_meshgrid`.

        """
        pix = self.pixel_shape.to(dtype=AP_config.ap_dtype)
        xsteps = (
            0.5
            * torch.arange(
                2 * (pix[0]) + 1,
                dtype=AP_config.ap_dtype,
                device=AP_config.ap_device,
            )
            - 0.5
        )
        ysteps = (
            0.5
            * torch.arange(
                2 * (pix[1]) + 1,
                dtype=AP_config.ap_dtype,
                device=AP_config.ap_device,
            )
            - 0.5
        )
        meshx, meshy = torch.meshgrid(
            xsteps,
            ysteps,
            indexing="xy",
        )
        Coords = self.pixel_to_plane(meshx, meshy)
        return torch.stack(Coords)

    # Window Comparison operators
    @torch.no_grad()
    def __eq__(self, other):
        return (
            torch.all(self.pixel_shape == other.pixel_shape)
            and torch.all(self.pixelscale == other.pixelscale)
            and (self.projection == other.projection)
            and (
                torch.all(
                    self.pixel_to_plane(torch.zeros_like(self.reference_imageij))
                    == other.pixel_to_plane(torch.zeros_like(other.reference_imageij))
                )
            )
        )  # fixme more checks?

    @torch.no_grad()
    def __ne__(self, other):
        return not self == other

    # Window interaction operators
    @torch.no_grad()
    def __or__(self, other):
        other_origin_pix = self.plane_to_pixel(other.origin)
        new_origin_pix = torch.minimum(-0.5 * torch.ones_like(other_origin_pix), other_origin_pix)

        other_pixel_end = self.plane_to_pixel(other.origin + other.end)
        new_pixel_end = torch.maximum(
            self.pixel_shape.to(dtype=AP_config.ap_dtype), other_pixel_end
        )
        return self.copy(
            origin=self.pixel_to_plane(new_origin_pix),
            pixel_shape=new_pixel_end - new_origin_pix,
        )

    @torch.no_grad()
    def __ior__(self, other):
        other_origin_pix = self.plane_to_pixel(other.origin)
        new_origin_pix = torch.minimum(-0.5 * torch.ones_like(other_origin_pix), other_origin_pix)

        other_pixel_end = self.plane_to_pixel(other.origin + other.end)
        new_pixel_end = torch.maximum(
            self.pixel_shape.to(dtype=AP_config.ap_dtype), other_pixel_end
        )

        self.reference_imageij = self.reference_imageij - (new_origin_pix + 0.5)
        self.pixel_shape = new_pixel_end - new_origin_pix
        return self

    @torch.no_grad()
    def __and__(self, other):
        other_origin_pix = self.plane_to_pixel(other.origin)
        new_origin_pix = torch.maximum(-0.5 * torch.ones_like(other_origin_pix), other_origin_pix)

        other_pixel_end = self.plane_to_pixel(other.origin + other.end)
        new_pixel_end = torch.minimum(
            self.pixel_shape.to(dtype=AP_config.ap_dtype) - 0.5, other_pixel_end
        )
        return self.copy(
            origin=self.pixel_to_plane(new_origin_pix),
            pixel_shape=new_pixel_end - new_origin_pix,
        )

    @torch.no_grad()
    def __iand__(self, other):
        other_origin_pix = self.plane_to_pixel(other.origin)
        new_origin_pix = torch.maximum(-0.5 * torch.ones_like(other_origin_pix), other_origin_pix)

        other_pixel_end = self.plane_to_pixel(other.origin + other.end)
        new_pixel_end = torch.minimum(
            self.pixel_shape.to(dtype=AP_config.ap_dtype), other_pixel_end
        )

        self.reference_imageij = self.reference_imageij - (new_origin_pix + 0.5)
        self.pixel_shape = new_pixel_end - new_origin_pix
        return self

    def __str__(self):
        return f"window origin: {self.origin.detach().cpu().tolist()}, shape: {self.shape.detach().cpu().tolist()}, center: {self.center.detach().cpu().tolist()}, pixelscale: {self.pixelscale.detach().cpu().tolist()}"

    def __repr__(self):
        return (
            f"window pixel_shape: {self.pixel_shape.detach().cpu().tolist()}, shape: {self.shape.detach().cpu().tolist()}\n"
            + super().__repr__()
        )


class Window_List(Window):
    def __init__(self, window_list=None, state=None):
        if state is not None:
            self.set_state(state)
        else:
            if window_list is None:
                window_list = []
            self.window_list = list(window_list)

        self.check_wcs()

    def check_wcs(self):
        """Ensure the WCS systems being used by all the windows in this list
        are consistent with each other. They should all project world
        coordinates onto the same tangent plane.

        """
        windows = tuple(
            W.reference_radec for W in filter(lambda w: w is not None, self.window_list)
        )
        if len(windows) == 0:
            return
        ref = torch.stack(windows)
        if not torch.allclose(ref, ref[0]):
            raise ConflicingWCS(
                "Reference (world) coordinate mismatch! All windows in Window_List are not on the same tangent plane! Likely serious coordinate mismatch problems. See the coordinates page in the documentation for what this means."
            )

        ref = torch.stack(
            tuple(W.reference_planexy for W in filter(lambda w: w is not None, self.window_list))
        )
        if not torch.allclose(ref, ref[0]):
            raise ConflicingWCS(
                "Reference (tangent plane) coordinate mismatch! All windows in Window_List are not on the same tangent plane! Likely serious coordinate mismatch problems. See the coordinates page in the documentation for what this means."
            )

        if len(set(W.projection for W in filter(lambda w: w is not None, self.window_list))) > 1:
            raise ConflicingWCS(
                "Projection mismatch! All windows in Window_List are not on the same tangent plane! Likely serious coordinate mismatch problems. See the coordinates page in the documentation for what this means."
            )

    @property
    @torch.no_grad()
    def origin(self):
        return tuple(w.origin for w in self)

    @property
    @torch.no_grad()
    def shape(self):
        return tuple(w.shape for w in self)

    @property
    @torch.no_grad()
    def center(self):
        return tuple(w.center for w in self)

    def shift_origin(self, shift):
        raise NotImplementedError("shift origin not implemented for window list")

    def copy(self):
        return self.__class__(list(w.copy() for w in self))

    def to(self, dtype=None, device=None):
        if dtype is None:
            dtype = AP_config.ap_dtype
        if device is None:
            device = AP_config.ap_device
        for window in self:
            window.to(dtype, device)

    def get_state(self):
        return list(window.get_state() for window in self)

    def set_state(self, state):
        self.window_list = list(Window(state=st) for st in state)

    # Window interaction operators
    @torch.no_grad()
    def __or__(self, other):
        new_windows = list((sw | ow) for sw, ow in zip(self, other))
        return self.__class__(window_list=new_windows)

    @torch.no_grad()
    def __ior__(self, other):
        for sw, ow in zip(self, other):
            sw |= ow
        return self

    @torch.no_grad()
    def __and__(self, other):
        new_windows = list((sw & ow) for sw, ow in zip(self, other))
        return self.__class__(window_list=new_windows)

    @torch.no_grad()
    def __iand__(self, other):
        for sw, ow in zip(self, other):
            sw &= ow
        return self

    # Window Comparison operators
    @torch.no_grad()
    def __eq__(self, other):
        results = list((sw == ow).view(-1) for sw, ow in zip(self, other))
        return torch.all(torch.cat(results))

    @torch.no_grad()
    def __ne__(self, other):
        return not self == other

    def __len__(self):
        return len(self.window_list)

    def __iter__(self):
        return (win for win in self.window_list)

    def __str__(self):
        return "Window List: \n" + ("\n".join(list(str(window) for window in self)) + "\n")

    def __repr__(self):
        return "Window List: \n" + ("\n".join(list(repr(window) for window in self)) + "\n")



from .base import Node

__all__ = ("Param_Unlock", "Param_SoftLimits", "Param_Mask")


class Param_Unlock:
    """Temporarily unlock a parameter.

    Context manager to unlock a parameter temporarily. Inside the
    context, the parameter will behave as unlocked regardless of its
    initial condition. Upon exiting the context, the parameter will
    return to its previous locked state regardless of any changes
    made by the user to the lock state.

    """

    def __init__(self, param=None):
        self.param = param

    def __enter__(self):
        if self.param is None:
            Node.global_unlock = True
        else:
            self.original_locked = self.param.locked
            self.param.locked = False

    def __exit__(self, *args, **kwargs):
        if self.param is None:
            Node.global_unlock = False
        else:
            self.param.locked = self.original_locked


class Param_SoftLimits:
    """Temporarily allow writing parameter values outside limits.

    Values outside the limits will be quietly (no error/warning
    raised) shifted until they are within the boundaries of the
    parameter limits. Since the limits are non-inclusive, the soft
    limits will actually move a parameter by 0.001 into the parameter
    range. For example the axis ratio ``q`` has limits from (0,1) so
    if one were to write: ``q.value = 2`` then the actual value that
    gets written would be ``0.999``.

    Cyclic parameters are not affected by this, any value outside the
    range is always (Param_SoftLimits context or not) wrapped back
    into the range using modulo arithmetic.

    """

    def __init__(self, param):
        self.param = param

    def __enter__(self, *args, **kwargs):
        self.original_setter = self.param._set_val_self
        self.param._set_val_self = self.param._soft_set_val_self

    def __exit__(self, *args, **kwargs):
        self.param._set_val_self = self.original_setter


class Param_Mask:
    """Temporarily mask parameters.

    Select a subset of parameters to be used through the "vector"
    interface of the DAG. The context is initialized with a
    Parameter_Node object (``P``) and a torch tensor (``M``) where the
    size of the mask should be equal to the current vector
    representation of the parameter (``M.numel() ==
    P.vector_values().numel()``). The mask tensor should be of
    ``torch.bool`` dtype where ``True`` indicates to keep using that
    parameter and ``False`` indicates to hide that parameter value.

    Note that ``Param_Mask`` contexts can be nested and will behave
    accordingly (the mask tensor will need to match the vector size
    within the previous context). As an example, imagine there is a
    parameter node ``P`` which has five sub-nodes each with a single
    value, one could nest contexts like::

      M1 = torch.tensor((1,1,0,1,0), dtype = torch.bool)
      with Param_Mask(P, M1):
        # Now P behaves as if it only has 3 elements
        M2 = torch.tensor([0,1,1], dtype = torch.bool)
        with Param_Mask(P, M2):
          # Now P behaves as if it only has 2 elements
          P.vector_values() # returns tensor with 2 elements

    """

    def __init__(self, param, new_mask):
        self.param = param
        self.new_mask = new_mask

    def __enter__(self):

        self.old_mask = self.param.vector_mask()
        self.mask = self.param.vector_mask()
        self.mask[self.mask.clone()] = self.new_mask
        self.param.vector_set_mask(self.mask)

    def __exit__(self, *args, **kwargs):
        self.param.vector_set_mask(self.old_mask)



from functools import lru_cache

import numpy as np
import torch
from astropy.convolution import convolve_fft

from .operations import fft_convolve_torch


def _h_poly(t):
    """Helper function to compute the 'h' polynomial matrix used in the
    cubic spline.

    Args:
        t (Tensor): A 1D tensor representing the normalized x values.

    Returns:
        Tensor: A 2D tensor of size (4, len(t)) representing the 'h' polynomial matrix.

    """

    tt = t[None, :] ** (torch.arange(4, device=t.device)[:, None])
    A = torch.tensor(
        [[1, 0, -3, 2], [0, 1, -2, 1], [0, 0, 3, -2], [0, 0, -1, 1]],
        dtype=t.dtype,
        device=t.device,
    )
    return A @ tt


def cubic_spline_torch(
    x: torch.Tensor, y: torch.Tensor, xs: torch.Tensor, extend: str = "const"
) -> torch.Tensor:
    """Compute the 1D cubic spline interpolation for the given data points
    using PyTorch.

    Args:
        x (Tensor): A 1D tensor representing the x-coordinates of the known data points.
        y (Tensor): A 1D tensor representing the y-coordinates of the known data points.
        xs (Tensor): A 1D tensor representing the x-coordinates of the positions where
                     the cubic spline function should be evaluated.
        extend (str, optional): The method for handling extrapolation, either "const" or "linear".
                                Default is "const".
                                "const": Use the value of the last known data point for extrapolation.
                                "linear": Use linear extrapolation based on the last two known data points.

    Returns:
        Tensor: A 1D tensor representing the interpolated values at the specified positions (xs).

    """
    m = (y[1:] - y[:-1]) / (x[1:] - x[:-1])
    m = torch.cat([m[[0]], (m[1:] + m[:-1]) / 2, m[[-1]]])
    idxs = torch.searchsorted(x[:-1], xs) - 1
    dx = x[idxs + 1] - x[idxs]
    hh = _h_poly((xs - x[idxs]) / dx)
    ret = hh[0] * y[idxs] + hh[1] * m[idxs] * dx + hh[2] * y[idxs + 1] + hh[3] * m[idxs + 1] * dx
    if extend == "const":
        ret[xs > x[-1]] = y[-1]
    elif extend == "linear":
        indices = xs > x[-1]
        ret[indices] = y[-1] + (xs[indices] - x[-1]) * (y[-1] - y[-2]) / (x[-1] - x[-2])
    return ret


def interpolate_bicubic(img, X, Y):
    """
    wrapper for scipy bivariate spline interpolation
    """
    f_interp = RectBivariateSpline(
        np.arange(dat.shape[0], dtype=np.float32),
        np.arange(dat.shape[1], dtype=np.float32),
        dat,
    )
    return f_interp(Y, X, grid=False)


def Lanczos_kernel_np(dx, dy, scale):
    """convolution kernel for shifting all pixels in a grid by some
    sub-pixel length.

    """
    xx = np.arange(-scale, scale + 1) - dx
    if dx < 0:
        xx *= -1
    Lx = np.sinc(xx) * np.sinc(xx / scale)
    if dx > 0:
        Lx[0] = 0
    else:
        Lx[-1] = 0

    yy = np.arange(-scale, scale + 1) - dy
    if dy < 0:
        yy *= -1
    Ly = np.sinc(yy) * np.sinc(yy / scale)
    if dx > 0:
        Ly[0] = 0
    else:
        Ly[-1] = 0

    LXX, LYY = np.meshgrid(Lx, Ly, indexing="xy")
    LL = LXX * LYY
    w = np.sum(LL)
    LL /= w
    # plt.imshow(LL.detach().numpy(), origin = "lower")
    # plt.show()
    return LL


def Lanczos_kernel(dx, dy, scale):
    """Kernel function for Lanczos interpolation, defines the
    interpolation behavior between pixels.

    """
    xx = np.arange(-scale + 1, scale + 1) + dx
    yy = np.arange(-scale + 1, scale + 1) + dy
    Lx = np.sinc(xx) * np.sinc(xx / scale)
    Ly = np.sinc(yy) * np.sinc(yy / scale)
    LXX, LYY = np.meshgrid(Lx, Ly)
    LL = LXX * LYY
    w = np.sum(LL)
    LL /= w
    return LL


def point_Lanczos(I, X, Y, scale):
    """
    Apply Lanczos interpolation to evaluate a single point.
    """
    ranges = [
        [int(np.floor(X) - scale + 1), int(np.floor(X) + scale + 1)],
        [int(np.floor(Y) - scale + 1), int(np.floor(Y) + scale + 1)],
    ]
    LL = Lanczos_kernel(np.floor(X) - X, np.floor(Y) - Y, scale)
    LL = LL[
        max(0, -ranges[1][0]) : LL.shape[0] + min(0, I.shape[0] - ranges[1][1]),
        max(0, -ranges[0][0]) : LL.shape[1] + min(0, I.shape[1] - ranges[0][1]),
    ]
    F = I[
        max(0, ranges[1][0]) : min(I.shape[0], ranges[1][1]),
        max(0, ranges[0][0]) : min(I.shape[1], ranges[0][1]),
    ]
    return np.sum(F * LL)


def _shift_Lanczos_kernel_torch(dx, dy, scale, dtype, device):
    """convolution kernel for shifting all pixels in a grid by some
    sub-pixel length.

    """
    xsign = 1 - 2 * (dx < 0).to(dtype=torch.int32)  # flips the kernel if the shift is negative
    xx = xsign * (torch.arange(int(-scale), int(scale + 1), dtype=dtype, device=device) - dx)
    Lx = torch.sinc(xx) * torch.sinc(xx / scale)

    ysign = 1 - 2 * (dy < 0).to(dtype=torch.int32)
    yy = ysign * (torch.arange(int(-scale), int(scale + 1), dtype=dtype, device=device) - dy)
    Ly = torch.sinc(yy) * torch.sinc(yy / scale)

    LXX, LYY = torch.meshgrid(Lx, Ly, indexing="xy")
    LL = LXX * LYY
    w = torch.sum(LL)
    # plt.imshow(LL.detach().numpy(), origin = "lower")
    # plt.show()
    return LL / w


def shift_Lanczos_torch(I, dx, dy, scale, dtype, device, img_prepadded=True):
    """Apply Lanczos interpolation to shift by less than a pixel in x and
    y.

    """
    LL = _shift_Lanczos_kernel_torch(dx, dy, scale, dtype, device)
    ret = fft_convolve_torch(I, LL, img_prepadded=img_prepadded)
    return ret


def shift_Lanczos_np(I, dx, dy, scale):
    """Apply Lanczos interpolation to shift by less than a pixel in x and
    y.

    I: the image
    dx: amount by which the grid will be moved in the x-axis (the "data" is fixed and the grid moves). Should be a value from (-0.5,0.5)
    dy: amount by which the grid will be moved in the y-axis (the "data" is fixed and the grid moves). Should be a value from (-0.5,0.5)
    scale: dictates size of the Lanczos kernel. Full kernel size is 2*scale+1
    """
    LL = Lanczos_kernel_np(dx, dy, scale)
    return convolve_fft(I, LL, boundary="fill")


def interpolate_Lanczos_grid(img, X, Y, scale):
    """
    Perform Lanczos interpolation at a grid of points.
    https://pixinsight.com/doc/docs/InterpolationAlgorithms/InterpolationAlgorithms.html
    """

    sinc_X = list(
        np.sinc(np.arange(-scale + 1, scale + 1) - X[i] + np.floor(X[i]))
        * np.sinc((np.arange(-scale + 1, scale + 1) - X[i] + np.floor(X[i])) / scale)
        for i in range(len(X))
    )
    sinc_Y = list(
        np.sinc(np.arange(-scale + 1, scale + 1) - Y[i] + np.floor(Y[i]))
        * np.sinc((np.arange(-scale + 1, scale + 1) - Y[i] + np.floor(Y[i])) / scale)
        for i in range(len(Y))
    )

    # Extract an image which has the required dimensions
    use_img = np.take(
        np.take(
            img,
            np.arange(int(np.floor(Y[0]) - step + 1), int(np.floor(Y[-1]) + step + 1)),
            0,
            mode="clip",
        ),
        np.arange(int(np.floor(X[0]) - step + 1), int(np.floor(X[-1]) + step + 1)),
        1,
        mode="clip",
    )

    # Create a sliding window view of the image with the dimensions of the lanczos scale grid
    # window = np.lib.stride_tricks.sliding_window_view(use_img, (2*scale, 2*scale))

    # fixme going to need some broadcasting magic
    XX = np.ones((2 * scale, 2 * scale))
    res = np.zeros((len(Y), len(X)))
    for x, lowx, highx in zip(range(len(X)), np.floor(X) - step + 1, np.floor(X) + step + 1):
        for y, lowy, highy in zip(range(len(Y)), np.floor(Y) - step + 1, np.floor(Y) + step + 1):
            L = XX * sinc_X[x] * sinc_Y[y].reshape((sinc_Y[y].size, -1))
            res[y, x] = np.sum(use_img[lowy:highy, lowx:highx] * L) / np.sum(L)
    return res


def interpolate_Lanczos(img, X, Y, scale):
    """
    Perform Lanczos interpolation on an image at a series of specified points.
    https://pixinsight.com/doc/docs/InterpolationAlgorithms/InterpolationAlgorithms.html
    """
    flux = []

    for i in range(len(X)):
        box = [
            [
                max(0, int(round(np.floor(X[i]) - scale + 1))),
                min(img.shape[1], int(round(np.floor(X[i]) + scale + 1))),
            ],
            [
                max(0, int(round(np.floor(Y[i]) - scale + 1))),
                min(img.shape[0], int(round(np.floor(Y[i]) + scale + 1))),
            ],
        ]
        chunk = img[box[1][0] : box[1][1], box[0][0] : box[0][1]]
        XX = np.ones(chunk.shape)
        Lx = (
            np.sinc(np.arange(-scale + 1, scale + 1) - X[i] + np.floor(X[i]))
            * np.sinc((np.arange(-scale + 1, scale + 1) - X[i] + np.floor(X[i])) / scale)
        )[
            box[0][0]
            - int(round(np.floor(X[i]) - scale + 1)) : 2 * scale
            + box[0][1]
            - int(round(np.floor(X[i]) + scale + 1))
        ]
        Ly = (
            np.sinc(np.arange(-scale + 1, scale + 1) - Y[i] + np.floor(Y[i]))
            * np.sinc((np.arange(-scale + 1, scale + 1) - Y[i] + np.floor(Y[i])) / scale)
        )[
            box[1][0]
            - int(round(np.floor(Y[i]) - scale + 1)) : 2 * scale
            + box[1][1]
            - int(round(np.floor(Y[i]) + scale + 1))
        ]
        L = XX * Lx * Ly.reshape((Ly.size, -1))
        w = np.sum(L)
        flux.append(np.sum(chunk * L) / w)
    return np.array(flux)


def interp1d_torch(x_in, y_in, x_out):
    indices = torch.searchsorted(x_in[:-1], x_out) - 1
    weights = (y_in[1:] - y_in[:-1]) / (x_in[1:] - x_in[:-1])
    return y_in[indices] + weights[indices] * (x_out - x_in[indices])


def interp2d(
    im: torch.Tensor,
    x: torch.Tensor,
    y: torch.Tensor,
) -> torch.Tensor:
    """
    Interpolates a 2D image at specified coordinates.
    Similar to `torch.nn.functional.grid_sample` with `align_corners=False`.

    Args:
        im (Tensor): A 2D tensor representing the image.
        x (Tensor): A tensor of x coordinates (in pixel space) at which to interpolate.
        y (Tensor): A tensor of y coordinates (in pixel space) at which to interpolate.

    Returns:
        Tensor: Tensor with the same shape as `x` and `y` containing the interpolated values.
    """

    # Convert coordinates to pixel indices
    h, w = im.shape

    # reshape for indexing purposes
    start_shape = x.shape
    x = x.view(-1)
    y = y.view(-1)

    x0 = x.floor().long()
    y0 = y.floor().long()
    x1 = x0 + 1
    y1 = y0 + 1
    x0 = x0.clamp(0, w - 2)
    x1 = x1.clamp(1, w - 1)
    y0 = y0.clamp(0, h - 2)
    y1 = y1.clamp(1, h - 1)

    fa = im[y0, x0]
    fb = im[y1, x0]
    fc = im[y0, x1]
    fd = im[y1, x1]

    wa = (x1 - x) * (y1 - y)
    wb = (x1 - x) * (y - y0)
    wc = (x - x0) * (y1 - y)
    wd = (x - x0) * (y - y0)

    result = fa * wa + fb * wb + fc * wc + fd * wd

    return result.view(*start_shape)


@lru_cache(maxsize=32)
def curvature_kernel(dtype, device):
    kernel = torch.tensor(
        [
            [0.0, 1.0, 0.0],
            [1.0, -4, 1.0],
            [0.0, 1.0, 0.0],
        ],  # [[1., -2.0, 1.], [-2.0, 4, -2.0], [1.0, -2.0, 1.0]],
        device=device,
        dtype=dtype,
    )
    return kernel


@lru_cache(maxsize=32)
def simpsons_kernel(dtype, device):
    kernel = torch.ones(1, 1, 3, 3, dtype=dtype, device=device)
    kernel[0, 0, 1, 1] = 16.0
    kernel[0, 0, 1, 0] = 4.0
    kernel[0, 0, 0, 1] = 4.0
    kernel[0, 0, 1, 2] = 4.0
    kernel[0, 0, 2, 1] = 4.0
    kernel = kernel / 36.0
    return kernel



import torch

from .galaxy_model_object import Galaxy_Model
from .warp_model import Warp_Galaxy
from .superellipse_model import SuperEllipse_Galaxy, SuperEllipse_Warp
from .foureirellipse_model import FourierEllipse_Galaxy, FourierEllipse_Warp
from .ray_model import Ray_Galaxy
from .wedge_model import Wedge_Galaxy
from .psf_model_object import PSF_Model
from ._shared_methods import (
    parametric_initialize,
    parametric_segment_initialize,
    select_target,
)
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..utils.parametric_profiles import gaussian_np

__all__ = [
    "Gaussian_Galaxy",
    "Gaussian_SuperEllipse",
    "Gaussian_SuperEllipse_Warp",
    "Gaussian_FourierEllipse",
    "Gaussian_FourierEllipse_Warp",
    "Gaussian_Warp",
    "Gaussian_PSF",
]


def _x0_func(model_params, R, F):
    return R[4], F[0]


def _wrap_gauss(R, sig, flu):
    return gaussian_np(R, sig, 10**flu)


class Gaussian_Galaxy(Galaxy_Model):
    """Basic galaxy model with Gaussian as the radial light profile. The
    gaussian radial profile is defined as:

    I(R) = F * exp(-0.5 R^2/S^2) / sqrt(2pi*S^2)

    where I(R) is the prightness as a function of semi-major axis
    length, F is the total flux in the model, R is the semi-major
    axis, and S is the standard deviation.

    Parameters:
        sigma: standard deviation of the gaussian profile, must be a positive value
        flux: the total flux in the gaussian model, represented as the log of the total

    """

    model_type = f"gaussian {Galaxy_Model.model_type}"
    parameter_specs = {
        "sigma": {"units": "arcsec", "limits": (0, None)},
        "flux": {"units": "log10(flux)"},
    }
    _parameter_order = Galaxy_Model._parameter_order + ("sigma", "flux")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_gauss, ("sigma", "flux"), _x0_func)

    from ._shared_methods import gaussian_radial_model as radial_model


class Gaussian_SuperEllipse(SuperEllipse_Galaxy):
    """Super ellipse galaxy model with Gaussian as the radial light
    profile.The gaussian radial profile is defined as:

    I(R) = F * exp(-0.5 R^2/S^2) / sqrt(2pi*S^2)

    where I(R) is the prightness as a function of semi-major axis
    length, F is the total flux in the model, R is the semi-major
    axis, and S is the standard deviation.

    Parameters:
        sigma: standard deviation of the gaussian profile, must be a positive value
        flux: the total flux in the gaussian model, represented as the log of the total

    """

    model_type = f"gaussian {SuperEllipse_Galaxy.model_type}"
    parameter_specs = {
        "sigma": {"units": "arcsec", "limits": (0, None)},
        "flux": {"units": "log10(flux)"},
    }
    _parameter_order = SuperEllipse_Galaxy._parameter_order + ("sigma", "flux")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_gauss, ("sigma", "flux"), _x0_func)

    from ._shared_methods import gaussian_radial_model as radial_model


class Gaussian_SuperEllipse_Warp(SuperEllipse_Warp):
    """super ellipse warp galaxy model with a gaussian profile for the
    radial light profile. The gaussian radial profile is defined as:

    I(R) = F * exp(-0.5 R^2/S^2) / sqrt(2pi*S^2)

    where I(R) is the prightness as a function of semi-major axis
    length, F is the total flux in the model, R is the semi-major
    axis, and S is the standard deviation.

    Parameters:
        sigma: standard deviation of the gaussian profile, must be a positive value
        flux: the total flux in the gaussian model, represented as the log of the total

    """

    model_type = f"gaussian {SuperEllipse_Warp.model_type}"
    parameter_specs = {
        "sigma": {"units": "arcsec", "limits": (0, None)},
        "flux": {"units": "log10(flux)"},
    }
    _parameter_order = SuperEllipse_Warp._parameter_order + ("sigma", "flux")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_gauss, ("sigma", "flux"), _x0_func)

    from ._shared_methods import gaussian_radial_model as radial_model


class Gaussian_FourierEllipse(FourierEllipse_Galaxy):
    """fourier mode perturbations to ellipse galaxy model with a gaussian
    profile for the radial light profile. The gaussian radial profile
    is defined as:

    I(R) = F * exp(-0.5 R^2/S^2) / sqrt(2pi*S^2)

    where I(R) is the prightness as a function of semi-major axis
    length, F is the total flux in the model, R is the semi-major
    axis, and S is the standard deviation.

    Parameters:
        sigma: standard deviation of the gaussian profile, must be a positive value
        flux: the total flux in the gaussian model, represented as the log of the total

    """

    model_type = f"gaussian {FourierEllipse_Galaxy.model_type}"
    parameter_specs = {
        "sigma": {"units": "arcsec", "limits": (0, None)},
        "flux": {"units": "log10(flux)"},
    }
    _parameter_order = FourierEllipse_Galaxy._parameter_order + ("sigma", "flux")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_gauss, ("sigma", "flux"), _x0_func)

    from ._shared_methods import gaussian_radial_model as radial_model


class Gaussian_FourierEllipse_Warp(FourierEllipse_Warp):
    """fourier mode perturbations to ellipse galaxy model with a gaussian
    profile for the radial light profile. The gaussian radial profile
    is defined as:

    I(R) = F * exp(-0.5 R^2/S^2) / sqrt(2pi*S^2)

    where I(R) is the prightness as a function of semi-major axis
    length, F is the total flux in the model, R is the semi-major
    axis, and S is the standard deviation.

    Parameters:
        sigma: standard deviation of the gaussian profile, must be a positive value
        flux: the total flux in the gaussian model, represented as the log of the total

    """

    model_type = f"gaussian {FourierEllipse_Warp.model_type}"
    parameter_specs = {
        "sigma": {"units": "arcsec", "limits": (0, None)},
        "flux": {"units": "log10(flux)"},
    }
    _parameter_order = FourierEllipse_Warp._parameter_order + ("sigma", "flux")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_gauss, ("sigma", "flux"), _x0_func)

    from ._shared_methods import gaussian_radial_model as radial_model


class Gaussian_Warp(Warp_Galaxy):
    """Coordinate warped galaxy model with Gaussian as the radial light
    profile. The gaussian radial profile is defined as:

    I(R) = F * exp(-0.5 R^2/S^2) / sqrt(2pi*S^2)

    where I(R) is the prightness as a function of semi-major axis
    length, F is the total flux in the model, R is the semi-major
    axis, and S is the standard deviation.

    Parameters:
        sigma: standard deviation of the gaussian profile, must be a positive value
        flux: the total flux in the gaussian model, represented as the log of the total

    """

    model_type = f"gaussian {Warp_Galaxy.model_type}"
    parameter_specs = {
        "sigma": {"units": "arcsec", "limits": (0, None)},
        "flux": {"units": "log10(flux)"},
    }
    _parameter_order = Warp_Galaxy._parameter_order + ("sigma", "flux")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_gauss, ("sigma", "flux"), _x0_func)

    from ._shared_methods import gaussian_radial_model as radial_model


class Gaussian_PSF(PSF_Model):
    """Basic point source model with a Gaussian as the radial light profile. The
    gaussian radial profile is defined as:

    I(R) = F * exp(-0.5 R^2/S^2) / sqrt(2pi*S^2)

    where I(R) is the prightness as a function of semi-major axis
    length, F is the total flux in the model, R is the semi-major
    axis, and S is the standard deviation.

    Parameters:
        sigma: standard deviation of the gaussian profile, must be a positive value
        flux: the total flux in the gaussian model, represented as the log of the total

    """

    model_type = f"gaussian {PSF_Model.model_type}"
    parameter_specs = {
        "sigma": {"units": "arcsec", "limits": (0, None)},
        "flux": {"units": "log10(flux)", "value": 0.0, "locked": True},
    }
    _parameter_order = PSF_Model._parameter_order + ("sigma", "flux")
    usable = True
    model_integrated = False

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_gauss, ("sigma", "flux"), _x0_func)

    from ._shared_methods import gaussian_radial_model as radial_model
    from ._shared_methods import radial_evaluate_model as evaluate_model


class Gaussian_Ray(Ray_Galaxy):
    """ray galaxy model with a gaussian profile for the radial light
    model. The gaussian radial profile is defined as:

    I(R) = F * exp(-0.5 R^2/S^2) / sqrt(2pi*S^2)

    where I(R) is the prightness as a function of semi-major axis
    length, F is the total flux in the model, R is the semi-major
    axis, and S is the standard deviation.

    Parameters:
        sigma: standard deviation of the gaussian profile, must be a positive value
        flux: the total flux in the gaussian model, represented as the log of the total

    """

    model_type = f"gaussian {Ray_Galaxy.model_type}"
    parameter_specs = {
        "sigma": {"units": "arcsec", "limits": (0, None)},
        "flux": {"units": "log10(flux)"},
    }
    _parameter_order = Ray_Galaxy._parameter_order + ("sigma", "flux")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_segment_initialize(
            model=self,
            parameters=parameters,
            target=target,
            prof_func=_wrap_gauss,
            params=("sigma", "flux"),
            x0_func=_x0_func,
            segments=self.rays,
        )

    from ._shared_methods import gaussian_iradial_model as iradial_model


class Gaussian_Wedge(Wedge_Galaxy):
    """wedge galaxy model with a gaussian profile for the radial light
    model. The gaussian radial profile is defined as:

    I(R) = F * exp(-0.5 R^2/S^2) / sqrt(2pi*S^2)

    where I(R) is the prightness as a function of semi-major axis
    length, F is the total flux in the model, R is the semi-major
    axis, and S is the standard deviation.

    Parameters:
        sigma: standard deviation of the gaussian profile, must be a positive value
        flux: the total flux in the gaussian model, represented as the log of the total

    """

    model_type = f"gaussian {Wedge_Galaxy.model_type}"
    parameter_specs = {
        "sigma": {"units": "arcsec", "limits": (0, None)},
        "flux": {"units": "log10(flux)"},
    }
    _parameter_order = Wedge_Galaxy._parameter_order + ("sigma", "flux")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_segment_initialize(
            self,
            parameters,
            target,
            _wrap_gauss,
            ("sigma", "flux"),
            _x0_func,
            self.wedges,
        )

    from ._shared_methods import gaussian_iradial_model as iradial_model



from typing import Optional

import torch

from .core_model import AstroPhot_Model
from ..image import (
    Image,
    Model_Image,
    Window,
    PSF_Image,
    Image_List,
)
from ._shared_methods import select_target
from ..utils.decorators import default_internal, ignore_numpy_warnings
from ..param import Parameter_Node
from ..errors import SpecificationConflict


__all__ = ["PSF_Model"]


class PSF_Model(AstroPhot_Model):
    """Prototype point source (typically a star) model, to be subclassed
    by other point source models which define specific behavior.

    PSF_Models behave differently than component models. For starters,
    their target image must be a PSF_Image object instead of a
    Target_Image object. PSF_Models also don't define a "center"
    variable since their center is always (0,0) just like a
    PSF_Image. A PSF_Model will never be convolved with a PSF_Model
    (that's it's job!), so a lot of the sampling method is simpler.

    """

    # Specifications for the model parameters including units, value, uncertainty, limits, locked, and cyclic
    parameter_specs = {
        "center": {
            "units": "arcsec",
            "value": (0.0, 0.0),
            "uncertainty": (0.1, 0.1),
            "locked": True,
        },
    }
    # Fixed order of parameters for all methods that interact with the list of parameters
    _parameter_order = ("center",)
    model_type = f"psf {AstroPhot_Model.model_type}"
    usable = False
    model_integrated = None

    # The sampled PSF will be normalized to a total flux of 1 within the window
    normalize_psf = True

    # Method for initial sampling of model
    sampling_mode = "simpsons"  # midpoint, trapezoid, simpson

    # Level to which each pixel should be evaluated
    sampling_tolerance = 1e-3

    # Integration scope for model
    integrate_mode = "threshold"  # none, threshold, full*

    # Maximum recursion depth when performing sub pixel integration
    integrate_max_depth = 3

    # Amount by which to subdivide pixels when doing recursive pixel integration
    integrate_gridding = 5

    # The initial quadrature level for sub pixel integration. Please always choose an odd number 3 or higher
    integrate_quad_level = 3

    # Maximum size of parameter list before jacobian will be broken into smaller chunks, this is helpful for limiting the memory requirements to build a model, lower jacobian_chunksize is slower but uses less memory
    jacobian_chunksize = 10
    image_chunksize = 1000

    # Softening length used for numerical stability and/or integration stability to avoid discontinuities (near R=0)
    softening = 1e-3

    # Parameters which are treated specially by the model object and should not be updated directly when initializing
    special_kwargs = ["parameters", "filename", "model_type"]
    track_attrs = [
        "sampling_mode",
        "sampling_tolerance",
        "integrate_mode",
        "integrate_max_depth",
        "integrate_gridding",
        "integrate_quad_level",
        "jacobian_chunksize",
        "softening",
    ]

    def __init__(self, *, name=None, **kwargs):
        self._target_identity = None
        super().__init__(name=name, **kwargs)

        # Set any user defined attributes for the model
        for kwarg in kwargs:  # fixme move to core model?
            # Skip parameters with special behaviour
            if kwarg in self.special_kwargs:
                continue
            # Set the model parameter
            setattr(self, kwarg, kwargs[kwarg])

        # If loading from a file, get model configuration then exit __init__
        if "filename" in kwargs:
            self.load(kwargs["filename"], new_name=name)
            return

        self.parameter_specs = self.build_parameter_specs(kwargs.get("parameters", None))
        with torch.no_grad():
            self.build_parameters()
            if isinstance(kwargs.get("parameters", None), torch.Tensor):
                self.parameters.value = kwargs["parameters"]
        assert torch.allclose(
            self.window.center, torch.zeros_like(self.window.center)
        ), "PSF models must always be centered at (0,0)"

    # Initialization functions
    ######################################################################
    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(
        self,
        target: Optional["PSF_Image"] = None,
        parameters: Optional[Parameter_Node] = None,
        **kwargs,
    ):
        """Determine initial values for the center coordinates. This is done
        with a local center of mass search which iterates by finding
        the center of light in a window, then iteratively updates
        until the iterations move by less than a pixel.

        Args:
          target (Optional[Target_Image]): A target image object to use as a reference when setting parameter values

        """
        super().initialize(target=target, parameters=parameters)

    @default_internal
    def transform_coordinates(self, X, Y, image=None, parameters=None):
        return X, Y

    # Fit loop functions
    ######################################################################
    def evaluate_model(
        self,
        X: Optional[torch.Tensor] = None,
        Y: Optional[torch.Tensor] = None,
        image: Optional[Image] = None,
        parameters: "Parameter_Node" = None,
        **kwargs,
    ):
        """Evaluate the model on every pixel in the given image. The
        basemodel object simply returns zeros, this function should be
        overloaded by subclasses.

        Args:
          image (Image): The image defining the set of pixels on which to evaluate the model

        """
        if X is None or Y is None:
            Coords = image.get_coordinate_meshgrid()
            X, Y = Coords - parameters["center"].value[..., None, None]
        return torch.zeros_like(X)  # do nothing in base model

    def make_model_image(self, window: Optional[Window] = None):
        """This is called to create a blank `Model_Image` object of the
        correct format for this model. This is typically used
        internally to construct the model image before filling the
        pixel values with the model.

        """
        if window is None:
            window = self.window
        else:
            window = self.window & window
        return self.target[window].blank_copy()

    def sample(
        self,
        image: Optional[Image] = None,
        window: Optional[Window] = None,
        parameters: Optional[Parameter_Node] = None,
    ):
        """Evaluate the model on the space covered by an image object. This
        function properly calls integration methods. This should not
        be overloaded except in special cases.

        This function is designed to compute the model on a given
        image or within a specified window. It takes care of sub-pixel
        sampling, recursive integration for high curvature regions,
        and proper alignment of the computed model with the original
        pixel grid. The final model is then added to the requested
        image.

        Args:
          image (Optional[Image]): An AstroPhot Image object (likely a Model_Image)
                                     on which to evaluate the model values. If not
                                     provided, a new Model_Image object will be created.
          window (Optional[Window]): A window within which to evaluate the model.
                                   Should only be used if a subset of the full image
                                   is needed. If not provided, the entire image will
                                   be used.

        Returns:
          Image: The image with the computed model values.

        """
        # Image on which to evaluate model
        if image is None:
            image = self.make_model_image(window=window)

        # Window within which to evaluate model
        if window is None:
            working_window = image.window.copy()
        else:
            working_window = window.copy()

        # Parameters with which to evaluate the model
        if parameters is None:
            parameters = self.parameters

        # Create an image to store pixel samples
        working_image = Model_Image(window=working_window)
        if self.model_integrated is True:
            # Evaluate the model on the image
            Coords = image.get_coordinate_meshgrid()
            X, Y = Coords - parameters["center"].value[..., None, None]
            working_image.data = self.evaluate_model(
                X=X, Y=Y, image=working_image, parameters=parameters
            )
        elif self.model_integrated is False:
            # Evaluate the model on the image
            reference, deep = self._sample_init(
                image=working_image,
                parameters=parameters,
                center=parameters["center"].value,
            )
            # Super-resolve and integrate where needed
            deep = self._sample_integrate(
                deep,
                reference,
                working_image,
                parameters,
                center=torch.zeros_like(working_image.center),
            )
            # Add the sampled/integrated pixels to the requested image
            working_image.data += deep
        else:
            raise SpecificationConflict(
                "PSF model 'model_integrated' should be either True or False"
            )

        # normalize to total flux 1
        if self.normalize_psf:
            working_image.data /= torch.sum(working_image.data)

        if self.mask is not None:
            working_image.data = working_image.data * torch.logical_not(self.mask)

        image += working_image

        return image

    @property
    def target(self):
        try:
            return self._target
        except AttributeError:
            return None

    @target.setter
    def target(self, tar):
        assert tar is None or isinstance(tar, PSF_Image)

        # If a target image list is assigned, pick out the target appropriate for this model
        if isinstance(tar, Image_List) and self._target_identity is not None:
            for subtar in tar:
                if subtar.identity == self._target_identity:
                    usetar = subtar
                    break
            else:
                raise KeyError(
                    f"Could not find target in Target_Image_List with matching identity to {self.name}: {self._target_identity}"
                )
        else:
            usetar = tar

        self._target = usetar

        # Remember the target identity to use
        try:
            self._target_identity = self._target.identity
        except AttributeError:
            pass

    def get_state(self, save_params=True):
        """Returns a dictionary with a record of the current state of the
        model.

        Specifically, the current parameter settings and the window for
        this model. From this information it is possible for the model to
        re-build itself lated when loading from disk. Note that the target
        image is not saved, this must be reset when loading the model.

        """
        state = super().get_state()
        state["window"] = self.window.get_state()
        if save_params:
            state["parameters"] = self.parameters.get_state()
        state["target_identity"] = self._target_identity
        for key in self.track_attrs:
            if getattr(self, key) != getattr(self.__class__, key):
                state[key] = getattr(self, key)
        return state

    # Extra background methods for the basemodel
    ######################################################################
    from ._model_methods import radius_metric
    from ._model_methods import angular_metric
    from ._model_methods import _sample_init
    from ._model_methods import _sample_integrate
    from ._model_methods import _integrate_reference
    from ._model_methods import build_parameter_specs
    from ._model_methods import build_parameters
    from ._model_methods import jacobian
    from ._model_methods import _chunk_jacobian
    from ._model_methods import _chunk_image_jacobian
    from ._model_methods import load



import torch

from .. import AP_config
from .image_object import Image, Image_List
from .window_object import Window
from ..utils.interpolate import shift_Lanczos_torch
from ..errors import InvalidImage

__all__ = ["Model_Image", "Model_Image_List"]


######################################################################
class Model_Image(Image):
    """Image object which represents the sampling of a model at the given
    coordinates of the image. Extra arithmetic operations are
    available which can update model values in the image. The whole
    model can be shifted by less than a pixel to account for sub-pixel
    accuracy.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target_identity = kwargs.get("target_identity", None)
        self.to()

    def clear_image(self):
        self.data = torch.zeros_like(self.data)

    def shift_origin(self, shift, is_prepadded=True):
        self.window.shift(shift)
        pix_shift = self.plane_to_pixel_delta(shift)
        if torch.any(torch.abs(pix_shift) > 1):
            raise NotImplementedError("Shifts larger than 1 pixel are currently not handled")
        self.data = shift_Lanczos_torch(
            self.data,
            pix_shift[0],
            pix_shift[1],
            min(min(self.data.shape), 10),
            dtype=AP_config.ap_dtype,
            device=AP_config.ap_device,
            img_prepadded=is_prepadded,
        )

    def get_window(self, window: Window, **kwargs):
        return super().get_window(window, target_identity=self.target_identity, **kwargs)

    def reduce(self, scale, **kwargs):
        return super().reduce(scale, target_identity=self.target_identity, **kwargs)

    def replace(self, other, data=None):
        if isinstance(other, Image):
            if self.window.overlap_frac(other.window) == 0.0:  # fixme control flow
                return
            other_indices = self.window.get_other_indices(other)
            self_indices = other.window.get_other_indices(self)
            if self.data[self_indices].nelement() == 0 or other.data[other_indices].nelement() == 0:
                return
            self.data[self_indices] = other.data[other_indices]
        elif isinstance(other, Window):
            self.data[self.window.get_self_indices(other)] = data
        else:
            self.data = other

    def get_state(self):
        state = super().get_state()
        state["target_identity"] = self.target_identity
        return state

    def set_state(self, state):
        super().set_state(state)
        self.target_identity = target_identity

    def get_fits_state(self):
        states = super().get_fits_state()
        for state in states:
            if state["HEADER"]["IMAGE"] == "PRIMARY":
                state["HEADER"]["TRGTID"] = self.target_identity
        return states

    def set_fits_state(self, states):
        super().set_fits_state(states)
        for state in states:
            if state["HEADER"]["IMAGE"] == "PRIMARY":
                self.target_identity = state["HEADER"]["TRGTID"]


######################################################################
class Model_Image_List(Image_List, Model_Image):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not all(isinstance(image, Model_Image) for image in self.image_list):
            raise InvalidImage(
                f"Model_Image_List can only hold Model_Image objects, not {tuple(type(image) for image in self.image_list)}"
            )

    def clear_image(self):
        for image in self.image_list:
            image.clear_image()

    def shift_origin(self, shift):
        raise NotImplementedError()

    def replace(self, other, data=None):
        if data is None:
            for image, oth in zip(self.image_list, other):
                image.replace(oth)
        else:
            for image, oth, dat in zip(self.image_list, other, data):
                image.replace(oth, dat)

    @property
    def target_identity(self):
        targets = tuple(image.target_identity for image in self.image_list)
        if any(tar_id is None for tar_id in targets):
            return None
        return targets

    def __isub__(self, other):
        if isinstance(other, Model_Image_List):
            for other_image, zip_self_image in zip(other.image_list, self.image_list):
                if other_image.target_identity is None or self.target_identity is None:
                    zip_self_image -= other_image
                    continue
                for self_image in self.image_list:
                    if other_image.target_identity == self_image.target_identity:
                        self_image -= other_image
                        break
                else:
                    self.image_list.append(other_image)
        elif isinstance(other, Model_Image):
            if other.target_identity is None or zip_self_image.target_identity is None:
                zip_self_image -= other_image
            else:
                for self_image in self.image_list:
                    if other.target_identity == self_image.target_identity:
                        self_image -= other
                        break
                else:
                    self.image_list.append(other)
        else:
            for self_image, other_image in zip(self.image_list, other):
                self_image -= other_image
        return self

    def __iadd__(self, other):
        if isinstance(other, Model_Image_List):
            for other_image, zip_self_image in zip(other.image_list, self.image_list):
                if other_image.target_identity is None or self.target_identity is None:
                    zip_self_image += other_image
                    continue
                for self_image in self.image_list:
                    if other_image.target_identity == self_image.target_identity:
                        self_image += other_image
                        break
                else:
                    self.image_list.append(other_image)
        elif isinstance(other, Model_Image):
            if other.target_identity is None or self.target_identity is None:
                for self_image in self.image_list:
                    self_image += other
            else:
                for self_image in self.image_list:
                    if other.target_identity == self_image.target_identity:
                        self_image += other
                        break
                else:
                    self.image_list.append(other)
        else:
            for self_image, other_image in zip(self.image_list, other):
                self_image += other_image
        return self
